# -*- coding: utf-8 -*-

import torchtext.vocab as vocab
import math
import numpy as np
import re
from sklearn.cluster import KMeans
from sklearn.svm import SVC
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.manifold import TSNE
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
import argparse
import logging
import os
import scipy

def reading(corpus_data_dir):
    with open(corpus_data_dir, "r") as fin:
        return fin.read()
def sentence_tokenize(raw_text):
    filters =  "\\【.*?】+|\\《.*?》+|\\#.*?#+|[/_$&%^*()<>+""'@|:~{}#]+|[——！\\\，。=？、：“”‘’￥……（）《》【】]"
    raw_sentence = raw_text
    cleanr = re.compile('<.*?>|[\\r]|[\\n]')
    raw_sentence = re.sub(cleanr, ' ', raw_sentence)
    raw_sentence = re.sub(filters,'',raw_sentence)
    return re.split("\.|,|!|\?|\"", raw_sentence)
def sentenceIndexList(word,word_to_index,sentence_matrix):
    sentence_index_list = []
    for i,sentence_vector in enumerate(sentence_matrix):
        if word in sentence_vector:
            sentence_index_list.append(i)
    return sentence_index_list
def norm_vector(raw_vector):
    norm_vector = []
    for eigenvalues in raw_vector:
        weight = eigenvalues / sum(raw_vector)
        norm_vector.append(weight)
    return norm_vector
def norm_matrix(matrix):
    row_max = matrix.max(axis=1)
    row_max = row_max.reshape(-1, 1)
    matrix = matrix - row_max
    norm_matrix = []
    for vector in matrix:
        norm_vector = []
        for eigenvalues in vector:
            x = eigenvalues / sum(vector)
            norm_vector.append(x)
        norm_matrix.append(norm_vector)
    return norm_matrix
def cos_similarity(x, y, norm=False):
    assert len(x) == len(y), "len(x) != len(y)"
    zero_list = [0] * len(x)
    if x == zero_list or y == zero_list:
        return float(1) if x == y else float(0)
    res = np.array([[x[i] * y[i], x[i] * x[i], y[i] * y[i]] for i in range(len(x))])
    cos = sum(res[:, 0]) / (np.sqrt(sum(res[:, 1])) * np.sqrt(sum(res[:, 2])))
    return 0.5 * cos + 0.5 if norm else cos
def remove_list(iterable: list, target: list):
        return [n for n in iterable if n not in target]
def transformMatrix(m):
    return list(zip(*m))
def sigmoid(x):
    return 1 / (1 + np.exp(-x))
def evalue_clustering(sentence_index_to_lable,sentence_matrix,sentence_index_list,k):
    for i in np.arange(k):
        print("Label:",sentence_index_to_lable[i]," | "," ".join(sentence_matrix[sentence_index_list[i]]))
def find_nearest_word(word,word_to_idx,embedding_weights):
    index = word_to_idx[word]
    embedding = embedding_weights[index]
    cos_dis = np.array([scipy.spatial.distance.cosine(e, embedding) for e in embedding_weights])
    return [idx_to_word[i] for i in cos_dis.argsort()[:10]]
def find_nearest_vec(vector,embedding_weights,idx_to_word):
    cos_dis = np.array([scipy.spatial.distance.cosine(e, vector) for e in embedding_weights])
    return [idx_to_word[i] for i in cos_dis.argsort()[:10]]

class meme_analysis(object):
    def __init__(self,pertrain_data_dir:str,corpus_data_dir:str,dimension:int=50):
        if not pertrain_data_dir:
            self.cache_dir = str(pertrain_data_dir)
            if not os.path.exists(dirs):
               os.makedirs(dirs)
        self.cache_dir = pertrain_data_dir
        self.dimension = dimension
        self.corpus_data_dir = corpus_data_dir
        self.glove,self.sentence_matrix,self.sentence_embedding_matrix = self._text_preprocessing(self.cache_dir,self.dimension,self.corpus_data_dir)
        
    # cache_dir : per-train data save（cache_dir = "./data"）
    # dimension : per-train data word embedding dimension (dimension = 50 or 100 or 200 or 300)
    # corpus_data_dir : Corpus for training morphemes (corpus_data = path of wikipedia_english)
    def _text_preprocessing(self,
                      cache_dir,
                      dimension,
                      corpus_data_dir
                      ):
        print("---sememe analysis start---")
        glove = vocab.pretrained_aliases["glove.6B."+str(dimension)+"d"](cache=cache_dir)
        print("number of vocabularies : ",len(glove.stoi))
        raw_text = reading(corpus_data_dir)
        print("corpus data preprocessing...")
        sentence = [w for w in sentence_tokenize(raw_text)]
        print("making sentence matrix...")
        sentence_matrix = [[lower_word for lower_word in w.lower().split()] for w in sentence]
        sentence_embedding_matrix = [[glove.stoi.get(word, "NON") for word in words]for words in sentence_matrix]
        sentence_embedding_matrix = [[i for i in sen if i != "NON"] for sen in sentence_embedding_matrix ]
        sentence_embedding_matrix = [sen for sen in sentence_embedding_matrix if len(sen)!=1 and len(sen)!=0]
        sentence_matrix = [[glove.itos[i] for i in sen_i] for sen_i in sentence_embedding_matrix]
        print("saved : model,sentence matrix and sentence embedding matrix")
        return glove,sentence_matrix,sentence_embedding_matrix
    def _calculating_morpheme_matrix(self,word,sentence_index_list,center_vector,index_to_vec,word_to_index,sentence_embedding_matrix):
        word_sememe_matrix = []
        for index in sentence_index_list:
            context_indices = remove_list(sentence_embedding_matrix[index],[word_to_index[word]])
            #context_indices = remove_list(context_indices,["NON"])
            if context_indices == []:
                print("context_indices == []")
            context_vec = [index_to_vec[i] for i in context_indices]
            weight_list = [cos_similarity(i,center_vector) for i in context_vec]
            if weight_list == []:
                print("weight_list == []")
            # len(weight_list)==len(context_vec)
            if len(weight_list) != len(context_vec):
                print("error : len(weight_list) != len(context_vec)")
            #if len(context_vec) == 1:
                #print("error : context_vec == 1 ; context_vec : ",[glove.itos[i] for i in context_indices])
            norm_weight_list = norm_vector(weight_list)
            weight_vector = np.multiply(np.array(norm_weight_list),np.array(transformMatrix(context_vec)))
            # len(weight_vector) == len(transformMatrix(context_vec))
            if len(weight_vector) != len(transformMatrix(context_vec)):
                print("error : len(weight_vector) != len(transformMatrix(context_vec))")
            relation_vector = [math.atan(sum(w)) for w in weight_vector]
            # len(relation_vector) == len(weight_vector)
            if len(relation_vector) != len(weight_vector):
                print("error : len(relation_vector) == len(weight_vector)")
            if relation_vector == []:
                print("error : relation_vector == [] | relation_vector: ",[glove.itos[i] for i in context_indices])
            word_sememe_matrix.append(relation_vector)
        return word_sememe_matrix
    
    def _num_of_sememe(self,max_num_of_sememe,dimensionality_reduction,view_detailed_clustering,sentence_index_list,word_sememe_matrix):
        for num_of_cluster in range(max_num_of_sememe):
            print("Trying to search the morphemes of word",num_of_cluster+2)
            cluster_center_matrix,sentence_index_to_lable_matrix,score_list= [],[],[]
            clustering = KMeans(num_of_cluster+2).fit(word_sememe_matrix)
            cluster_center_matrix.append(clustering.cluster_centers_)
            sentence_index_to_lable = clustering.labels_
            sentence_index_to_lable_matrix.append(clustering.labels_)
            if view_detailed_clustering:
                print("Text classification on morphemes <Top10>...")
                evalue_clustering(sentence_index_to_lable,sentence_index_list,40)
            if dimensionality_reduction:
                print("The cluster distribution scatter plot is being produced...")
                tsne = TSNE(perplexity=30, n_components=2, init='pca', n_iter=3000, method='exact')
                low_dim_embs = tsne.fit_transform(word_sememe_matrix)
                centers_low_dim_embs = tsne.fit_transform(clustering.cluster_centers_)
                plt.scatter(low_dim_embs[:, 0], low_dim_embs[:, 1], c=sentence_index_to_lable,s=50, cmap='viridis')
                print("------------------------")
            print("The classifier is being used to evaluate the clustering results...")
            x_train, x_test, y_train, y_test = train_test_split(word_sememe_matrix, sentence_index_to_lable, random_state=1, train_size=0.6)
            clf = make_pipeline(StandardScaler(), SVC(kernel='linear',gamma='auto'))
            clf.fit(x_train, y_train.ravel())
            print("Train score:",clf.score(x_train, y_train))
            print("Test score:",clf.score(x_test, y_test))
            score_list.append(clf.score(x_test, y_test))
        argmax_score = np.argmax(score_list)
        return cluster_center_matrix[argmax_score],sentence_index_to_lable_matrix[argmax_score]
        print("------------------------")
        
    def word2sememe(self,
                    word:str,
                    max_num_of_sememe:int=5,
                    dimensionality_reduction:bool=False,
                    view_detailed_clustering:bool=False
                    ):
        center_vector = self.glove.vectors[self.glove.stoi[word]]
        sentence_index_list = sentenceIndexList(word,self.glove.stoi,self.sentence_matrix)
        print("------------------------")
        print("Calculating morpheme matrix...")
        print("------------------------")
        word_sememe_matrix = self._calculating_morpheme_matrix(word,sentence_index_list,center_vector,self.glove.vectors,self.glove.stoi,self.sentence_embedding_matrix)
        print("------------------------")
        print("The morpheme matrix is completed!")
        print("------------------------")
        print("Trying to cluster the morpheme matrix...")
        print("------------------------")
        cluster_centers,sentence_index_to_lable = self._num_of_sememe(max_num_of_sememe,dimensionality_reduction,view_detailed_clustering,sentence_index_list,word_sememe_matrix)
        print("Text classification on morphemes <Top40>...")
        evalue_clustering(sentence_index_to_lable,self.sentence_matrix,sentence_index_list,40)
        print("------------------------")
        print("Find the word that is closest to the sum of phonemes...")
        sememe_sum = find_nearest_vec(sum(cluster_centers),self.glove.vectors,self.glove.itos)
        print(sememe_sum)
        print("------------------------")
        for sememe_i in cluster_centers:
            sememe_nearest = find_nearest_vec(sememe_i,self.glove.vectors,self.glove.itos)
            print("The closest word of the phoneme of the",word,":\n",sememe_nearest)
        print("------------------------End------------------------")
        return sememe_sum,cluster_centers
